import threading
import pickle as pk
import redis


from ..config import Config  # noqa

lock = threading.Lock()


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(
                        *args, **kwargs
                    )
        return cls._instances[cls]


class Redis(metaclass=Singleton):
    redis_instance = None

    def __init__(self):
        self.redis_instance = redis.StrictRedis(
                host=Config.config("REDIS_HOST"),
                port=Config.config("REDIS_PORT"),
                db=0,
                password=Config.config("REDIS_PASSWORD")
                if Config.config("REDIS_PASSWORD")
                else "",
            )
        print("Redis Initialized Successfully.")

    @staticmethod
    def cache(key, value, expire_in=None):
        val = pk.dumps(value)
        result = Redis().redis_instance.set(key, val)
        if expire_in:
            Redis().redis_instance.expire(key, expire_in)
        return result

    @staticmethod
    def acache(key, value, expire_in=None):
        cache_thread = threading.Thread(
            target=Redis().cache,
            args=(
                key,
                value,
            ),
            kwargs={"expire_in": expire_in},
        )
        cache_thread.start()

    @staticmethod
    def expire_in(key, seconds):
        Redis().redis_instance.expire(key, seconds)

    @staticmethod
    def get(key):
        if Redis().is_exist(key):
            val = Redis().redis_instance.get(key)
            value = pk.loads(val)
        else:
            value = None
        return value

    @staticmethod
    def is_exist(key):
        return True if Redis().redis_instance.exists(key) == 1 else False

    @staticmethod
    def delete(keys):
        for key in keys:
            Redis().redis_instance.delete(key)

    @staticmethod
    def deleteall():
        return Redis().redis_instance.flushall()

    @staticmethod
    def delete_current_db():
        return Redis().redis_instance.flushdb()

    @staticmethod
    def keys_count():
        return Redis().redis_instance.dbsize() - 1



import random

from math import sin, radians, cos, asin, sqrt
from .utils.type_hints import Location, LocationArray


# this is the direct distance between two points 
# on earth respecting the circularity of earth
def haversine_distance(
    location_a: Location, location_b: Location
) -> float:  # (source_lat, source_lng), (destination_lat, destination_lng)
    def hav(angle):
        return sin(angle / 2) ** 2

    # we convert from decimal degree to radians
    lat_a, lon_a, lat_b, lon_b = (
        radians(location_a[0]),
        radians(location_a[1]),
        radians(location_b[0]),
        radians(location_b[1]),
    )

    delta_lat = lat_b - lat_a
    delta_lon = lon_b - lon_a
    a = hav(delta_lat) + cos(lat_a) * cos(lat_b) * hav(delta_lon)
    c = 2 * asin(sqrt(a))
    # approximate radius of the Earth: 6371 km
    return c * 6371


def generate_random_locations(
    location: Location, number: int = 10, ratio: float = 50
) -> LocationArray:
    """
    Generate random locations:
        Parameters:
            - location: A tuple of str, it's a base location.
            - number: The number of needed random locations.
            - ratio: This decide how much the locations are far or close from 
            each others, higher ratio --> locations are closer.
        Returns:
            - locations: List of locations. 
    """
    return [
        (
            str(float(location[0]) + (random.random() / ratio)),
            str(float(location[1]) + (random.random() / ratio)),
        )
        for _ in range(number)
    ]



import os
from typing import Any
import yaml
from dotenv import load_dotenv

load_dotenv()


class Config:
    # You need to add this file to .gitignore of your project
    with open(
        os.getenv("CONFG_YAML_FILE_NAME", "osm_helper_config.yaml"), "r"
    ) as stream:
        try:
            OSM_CONFIG = yaml.safe_load(stream)
        except yaml.YAMLError as exc: # noqa
            OSM_CONFIG = {}
            pass

    __conf = OSM_CONFIG
    __conf.update(
        {
            "OSRM_URL": __conf.get(
                "PRIMARY_OSRM_URL",
                os.getenv(
                    "PRIMARY_OSRM_URL",
                    "http://router.project-osrm.org/",
                ),
            )
        }
    )
    __setters = ["OSRM_URL"]

    @staticmethod
    def config(name: str, default_value: Any=None) -> Any:
        return Config.__conf.get(name, default_value)

    @classmethod
    def set(cls, name: str, value: Any) -> None:
        if not isinstance(name, str):
            raise TypeError(f"{name} must be str not but {type(name)} is given.")

        if name in Config.__setters:
            cls.__conf[name] = value
        else:
            raise NameError(f'"{name}" not accepted in set() method')



import unittest

from osm_helper.examples.basic import default_parameters_example


class TestOSMApi(unittest.TestCase):
    def test_with_default_config(self):
        result, _ = default_parameters_example()
        assert 'duration' in result
        
        # since the number of locations in the default example is 10, so the result list is 5x5
        assert len(result["duration"]) == 5

        is_result_length_valid = True
        is_all_zero = True
        for item in result["duration"]:
            if len(item) != 5:
                is_result_length_valid = False
            
            for time in item:
                if time != 0:
                    is_all_zero = False
            
            if not is_result_length_valid and not is_all_zero:
                break

        assert is_result_length_valid
        assert not is_all_zero
